/*
   scalarprod_mutex
   


   EEE4084F prepated by S. Winberg 2018
*/

#include <pthread.h>
#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include "Timer.h"

/* pthreads_dotprod.cpp
   EEE4084F
   Parallelized pthreads example

*/

using namespace std;

// Define number of threads to use and length of vector
#define N_THREADS 6
 // amount of doubles each thread will process
//#define BLOCK_LEN 50000
#define BLOCK_LEN 33333

// Define data structure to be shared between the threads
struct DotprodData {
   double   *A;
   double   *B;
   double   sum;
   unsigned block_len;
 } ;

DotprodData     shared_data;       // data that is shared between the threads
pthread_t       thread[N_THREADS]; // thread data
pthread_mutex_t mux_sum;           // mutex needed for summing up the results

/** Thread operation, multiplied and adds one block of the shared vectors */
void *dotprod(void *arg)
{
   /* Define and use local variables for convenience */
   int i, m, start, stop;
   long my_id;
   double my_sum, *A, *B;
   long offset = (long)arg;

   // read in to local registers
   m     = shared_data.block_len;
   start = offset*m;
   stop  = start+m;
   // set up arrays
   A = shared_data.A;
   B = shared_data.B;

   // do the calculation
   my_sum = 0;
   for (i=start; i<stop; i++) {
       my_sum = my_sum + (A[i] * B[i]);
       }

   /* Lock mutex prior changuing the shared data. */
   pthread_mutex_lock (&mux_sum);
     /* This code is protected by the mutex mux_sum */
     shared_data.sum += my_sum;
     printf("thread#%lu computed %d -> %d:  my_sum=%f total_sum=%f\n",
            offset,start,stop,my_sum,shared_data.sum);
   pthread_mutex_unlock (&mux_sum);

   /* Exit the thread */
   pthread_exit(0);
}

/***********************************************************************
 main()  entry point to the program

 The main program creates threads which do all the work and then
 print out result upon completion. Before creating the threads,
 The input data is created. Since all threads update a shared structure, we
 need a mutex for mutual exclusion. The main thread needs to wait for
 all threads to complete, it waits for each one of the threads. We specify
 a thread attribute value that allow the main thread to join with the
 threads it creates. Note also that we free up handles  when they are
 no longer needed.

***********************************************************************/

int main (int argc, char *argv[])
{
   unsigned i;
   void     *pstatus;
   pthread_attr_t attr;

   // start the timer
   tic();

   /* Aollocate space and initialize shared data */
   shared_data.A = new double [N_THREADS*BLOCK_LEN];
      //alternatively: (double*) malloc (NUMTHRDS*VECLEN*sizeof(double));
   shared_data.B = new double [N_THREADS*BLOCK_LEN];
      //alternatively: (double*) malloc (NUMTHRDS*VECLEN*sizeof(double));
   shared_data.block_len = BLOCK_LEN;

   for (i=0; i<(BLOCK_LEN*N_THREADS); i++) {
      shared_data.A[i]=(double)(i*0.1);
      shared_data.B[i]=(double)(i*0.2);
      }
   /* set the sum to 0, note that the sum will be locked by a mutex */
   shared_data.sum=0;

   /* initialize the mutex for protecting the shared sum variable */
   pthread_mutex_init(&mux_sum, NULL);

   /* Initialize the threads library */
   pthread_attr_init(&attr);
   pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

   /* spawn the threads */
   for (i=0;i<N_THREADS;i++) {
       /* Spawn the threads. Each one works on a BLOCK_LEN segment of
        * the shared A,B vectors. */
       pthread_create(&thread[i], &attr, dotprod, (void *)(long)i);
       }

   /* can get rid of attributes sent to threads */
   pthread_attr_destroy(&attr);

   /* This main thread now just waits on the other threads to complete. */
   for (i=0;i<N_THREADS;i++) {
       pthread_join(thread[i], &pstatus);
       }

   /* At this point the joins are all complete, i.e. the threads have completed. */

  /* Display the results of the vector product... */
   printf ("Sum =  %f \n", shared_data.sum);
   delete shared_data.A, shared_data.B;

   /* Finally, don't forget to dispose the mutex */
   pthread_mutex_destroy(&mux_sum);

   /* Display the time taken */
   double t1 = toc();  // get final tick value
   cout << "scalarprod n=" << BLOCK_LEN*N_THREADS << " took " << t1 << " ms\n";

   /* This isn't really necessary but is best to do, tell the
      pthreads library to clean up and dispose of any memory. */
   pthread_exit(NULL);
}

